/*
   High Dynamic Range.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform high dynmaic range rendering.
   What this means is that we want to work with values that are above the
   normal 0 to 1 range (0 to 255).  In order to do this we will need a few
   things.  First we will need floating point textures.  The normal integer
   textures and surfaces will only allow us to use a certain range of
   values with low percision.  With the floating point buffers
   we can pass the 0 to 1 range all we want.  At this point if we rendered
   the scene to the screen, it would look like a super overexposed bright
   image.  To convert our HDR scene to a range that can be displayed we use
   tone mapping.  This will take our HDR values and make them LDR based
   on the average luminance in the scene.  Tone mapping in this demo
   is simple.  We simply multiply the color values by a exposure variable
   then use a power to the 0.55 to convert the HDR to LDR.
   
   The exposure level of the scene can be controlled by the up and down arrow
   keys.  The main object rendered in the first pass is just a small, simple
   quad with a HDR texture image.  This texture
   is a RGBE image.  A RGBE image is just a image with RGB data and a exponent
   (hence the E).  This exponenet is used to change the values stored in
   the RGB channels into HDR values.  Technically we don't need RGBE images
   to do HDR rendering but I used it to demonstrate how they can be used.
   You can download image probes from the internet or create your own with
   a program like HDR shop.  I must admit I just threw together my hdr
   image but you can find a whole lot nicer ones over the internet to
   test this program out on.
*/


// Demo: High Dynamic Range.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<windows.h>
#include<d3dx9.h>


// Our vertex stucture define.
#define D3DFVF_OBJVertex (D3DFVF_XYZ | D3DFVF_TEX1)
#define D3DFVF_ScreenVertex (D3DFVF_XYZRHW | D3DFVF_TEX1)

#define RELEASE_OBJ(a) if(a != NULL) { a->Release(); a = NULL; }

#define WINDOW_WIDTH 640
#define WINDOW_HEIGHT 480

// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);
bool InitializeObject();
void RenderScene();
bool LoadEffectShaders();
void ShutdownDirect3D();

struct Vertex
{
   float x, y, z;
   float tu, tv;
};

struct ScreenVertex
{
	float x, y, z, rhw;
	float tu, tv;
};


float fullOffset[] = { 1 / WINDOW_WIDTH, 1 / WINDOW_HEIGHT };

LPDIRECT3D9 Direct3D_Object = NULL;
LPDIRECT3DDEVICE9 D3D_Device = NULL;

// Main object and its texture.
LPDIRECT3DTEXTURE9 objectTexture = NULL;
LPDIRECT3DVERTEXBUFFER9 objectBuffer = NULL;

// Fullscreen size quad, half screen size quad.
LPDIRECT3DVERTEXBUFFER9 fullBuffer = NULL;

// Offscreen surfaces and textures.
LPDIRECT3DSURFACE9 backSurface = NULL;
LPDIRECT3DSURFACE9 fullSurface = NULL;
LPDIRECT3DTEXTURE9 fullTexture = NULL;

// Shaders.
LPD3DXEFFECT toneShader = NULL;

// Projection matrices.
D3DXMATRIXA16 fullProj;

// MVP matrix.
D3DXMATRIX worldViewProjection;

// Exposure level of the hdr image.
float exposure = 2.0f;


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   switch(message)
      {
         case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;

         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;

         default:
            break;
      }

    return DefWindowProc(hwnd, message, wParam, lParam);
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // true if we should exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class headerrmation.
   windowClass.cbWndExtra = 0;                        // Extra window headerrmation.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   hwnd = CreateWindowEx(NULL,               // The extended style.
                         "UGPClass",         // Window class.
                         "HDR Rendering - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |// Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |// Window style.
                          WS_CLIPSIBLINGS,   // Window style.
                          100, 100,          // X, Y coords.
                          WINDOW_WIDTH, WINDOW_HEIGHT,// Window size.
                          NULL,              // Handle to parent window.
                          NULL,              // Menu.
                          hInstance,         // Handle to app instance.
                          NULL);             // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   if(Direct3D_Object == NULL) return false;

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      return false;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = DisplayMode.Width;
         Present_Parameters.BackBufferHeight = DisplayMode.Height;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
         D3DCREATE_HARDWARE_VERTEXPROCESSING, &Present_Parameters, &D3D_Device)))
      {
         return false;
      }

   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);
   D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

   // Set the window perpective.
   D3DXMatrixPerspectiveFovLH(&fullProj, 45.0f, (float)(WINDOW_WIDTH / WINDOW_HEIGHT),
                              0.1f, 500.0f);
   D3D_Device->SetTransform(D3DTS_PROJECTION, &fullProj);

   // Initialize all render targets and textures.
   D3DXCreateTexture(D3D_Device, WINDOW_WIDTH, WINDOW_HEIGHT, 1, D3DUSAGE_RENDERTARGET,
                     D3DFMT_A16B16G16R16F, D3DPOOL_DEFAULT, &fullTexture);
	fullTexture->GetSurfaceLevel(0, &fullSurface);

   // Here we set up the object this demo will render.
   if(!InitializeObject()) return false;
   if(!LoadEffectShaders()) return false;

   return true;
}


bool InitializeObject()
{
   // Pointer to the buffer.
   Vertex *ptr;

   // Square data for its position and texture coords.
	Vertex obj[] =
	{
		{ -0.5f, 0.5f, 1.0f, 0.0, 0.0 },
      { 0.5f, 0.5f, 1.0f, 1.0, 0.0 },
	   { -0.5f, -0.5f, 1.0f, 0.0, 1.0 },
      { 0.5f, -0.5f, 1.0f, 1.0, 1.0 }
	};
	
	// Data for a fullscreen quad.
	ScreenVertex fullWin[] =
	{
		{ 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f },
      { 0.0f, WINDOW_HEIGHT, 0.0f, 1.0f, 0.0f, 1.0f },
	   { WINDOW_WIDTH, 0.0f, 0.0f, 1.0f, 1.0f, 0.0f },
      { WINDOW_WIDTH, WINDOW_HEIGHT, 0.0f, 1.0f, 1.0f, 1.0f }
	};

   // Create the vertex buffer that will hold the object.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(obj), 0, D3DFVF_OBJVertex,
                        D3DPOOL_DEFAULT, &objectBuffer, NULL))) return false;

   if(FAILED(objectBuffer->Lock(0, sizeof(obj), (void**)&ptr, 0))) return false;
   memcpy(ptr, obj, sizeof(obj));
   objectBuffer->Unlock();
   
   
   // Create the full screen buffer.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(fullWin), 0, D3DFVF_ScreenVertex,
                        D3DPOOL_DEFAULT, &fullBuffer, NULL))) return false;

   if(FAILED(fullBuffer->Lock(0, sizeof(fullWin), (void**)&ptr, 0))) return false;
   memcpy(ptr, fullWin, sizeof(fullWin));
   fullBuffer->Unlock();
   
   
   // Load the HDR image file.  D3DXCreateTextureFromFile() supports HDR images.
   if(D3DXCreateTextureFromFile(D3D_Device, "ugp.hdr", &objectTexture) != D3D_OK)
      return false;

   return true;
}


bool LoadEffectShaders()
{
   // We can use this to store any errors when compiling the shader effect file.
	LPD3DXBUFFER errs = NULL;

	if(FAILED(D3DXCreateEffectFromFile(D3D_Device, "tone.fx", NULL, NULL, 0,
                                      NULL, &toneShader, &errs)))
      {
         // Display any errors in message box.
		   LPVOID cErr = errs->GetBufferPointer();
         MessageBox(NULL, (const char*)cErr, "fx Error...", MB_OK | MB_ICONEXCLAMATION);
         return false;
	   }
	
	return true;
}


void FullScreenPass()
{
   // Set the render target to our offscreen surface.
   D3D_Device->SetRenderTarget(0, fullSurface);
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_COLORVALUE(0.0f, 0.0f, 0.0f, 1.0f), 1.0f, 0);

   D3D_Device->SetTransform(D3DTS_PROJECTION, &fullProj);

   D3D_Device->BeginScene();

      // Render object with HDR image on it.
      D3D_Device->SetTexture(0, objectTexture);
      D3D_Device->SetStreamSource(0, objectBuffer, 0, sizeof(Vertex));
      D3D_Device->SetFVF(D3DFVF_OBJVertex);
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

   D3D_Device->EndScene();
}


void TonePass()
{
   D3D_Device->SetRenderTarget(0, backSurface);
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_COLORVALUE(0.0f, 0.0f, 0.0f, 0.0f), 1.0f, 0);

   // Not necessary but this demo is a scaled down version of another
   // more advanced one I did so I will leave it for the heck of.
   worldViewProjection = fullProj;
   D3DXMatrixTranspose(&worldViewProjection, &worldViewProjection);

   D3D_Device->BeginScene();

      // Apply the shader.  The shader will need the MVP, fullscreen
      // surface as a texture, offset to alter the tex coords by to
      // get the right results, and a exposure variable to increase
      // or decrease the brightness of the scene.
      toneShader->SetTechnique("Tone");
      toneShader->SetMatrix("worldViewProj", &worldViewProjection);
      toneShader->SetTexture("fullTexture", fullTexture);
      toneShader->SetFloatArray("offset", fullOffset, 2);
      toneShader->SetFloat("exposure", exposure);

      // Apply the shader.
      toneShader->Begin(NULL, 0);
         toneShader->BeginPass(0);

            // Render fullscreen quad.
            D3D_Device->SetStreamSource(0, fullBuffer, 0, sizeof(ScreenVertex));
            D3D_Device->SetFVF(D3DFVF_ScreenVertex);
            D3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

         toneShader->EndPass();
      toneShader->End();

   D3D_Device->EndScene();
}


void RenderScene()
{
   // Get a copy of the back buffer.
   D3D_Device->GetBackBuffer(0, 0, D3DBACKBUFFER_TYPE_MONO, &backSurface);

   // Perform each pass.
   FullScreenPass();
   TonePass();

   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
   
   // Gather user input.
   if(GetKeyState(VK_DOWN) & 0x80) exposure -= 0.05f;
   if(GetKeyState(VK_UP) & 0x80) exposure += 0.05f;
   
   if(exposure < 0.0f) exposure = 0.01f;
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   RELEASE_OBJ(D3D_Device)
   RELEASE_OBJ(Direct3D_Object);
   RELEASE_OBJ(objectTexture);
   RELEASE_OBJ(objectBuffer);
   RELEASE_OBJ(fullBuffer);
   RELEASE_OBJ(toneShader);
   RELEASE_OBJ(backSurface);
   RELEASE_OBJ(fullSurface);
   RELEASE_OBJ(fullTexture);
}


// Recap:

// All high dynamic range rendering is is the ability to go over the normal 0 to 1 range.
// Once over this range since our monitors need values between 0 and 1 so we must use tone
// mapping to convert the values to that based on the average luminance.  A lot of things
// can be done with HDR rendering.  For example in the DirectX SDK there is a demo on HDR
// lighting that allows you to look into a dark area of the scene and your eyes will adjust
// just like it would in real life.  If you then look at a light source the scene will
// appear bright but your eyes will adjust to the lighting.  When I say eyes I don't mean
// your real eyes but the program will simulate this effect.

// During the first pass we just render the scene in an offscreen buffer just like
// we would if we were drawing to the back buffer like normal.  We then took this
// image and rendered it to a fullscreen quad.  This quad had a tone mapping shader
// applied to it that tone mapped our scene.  We could have just applied the tone
// mapper to the fullscreen pass and avoided rendering to a surface but this demo
// will be built on to allow HDR lighting with a bloom filter.  On top of that
// I plan on creating demos that does other effects with HDR and we will need
// to use offscreen buffers.  Better to add them now while we are at such a simple
// stage.


// Copyright March 2005
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com